def tauvalue(game):
    # Extract the set of players
    players = set([element for sublist in [i for i in game['coalition']] for element in sublist])

    # Change the data type of coalitions from list to tuple to be hashable
    game['coalition'] = [tuple(x) for x in game['coalition']]
    game = game.set_index('coalition')

    # Find the grand coalition
    gc_list = [coal for coal in game.index if len(coal) == len(players)][0]
    gc = game['value'][tuple(gc_list)]

    # Determine the marginal profits
    Mi = {}
    for player in players:
        Mi[player] = gc - game['value'][tuple([p for p in gc_list if p is not player])]

    # Determine the minimum claim
    mi = {}
    for player in players:
        mi[player] = max([game['value'][coal]-sum(Mi[p] for p in coal if p is not player) for coal in game.index])

    # Determine alpha
    alpha = (gc-sum(Mi[p] for p in players))/(sum(mi[p] for p in players)-sum(Mi[p] for p in players))

    # Determine tau
    tau={p: alpha*mi[p]+(1-alpha)*Mi[p] for p in players}
    print('Tau Value properly calculated\n')

    return(tau)


import pyomo.environ as pyo
from pyomo.opt import SolverStatus, TerminationCondition

def feasible_model():
    """
    Builds an abstract model of the utility exchange network.
    """
    model = pyo.AbstractModel()
    # ##########################
    # SETS
    # ##########################
    model.COAL = pyo.Set(doc='Set of coalitions')
    model.PLAYER = pyo.Set(doc='Set of players')
    model.PLAYER_SUB = pyo.Set(model.COAL, doc='Player subsets')
    # ##########################
    # VARIABLES
    # ##########################
    model.xj = pyo.Var(model.PLAYER, within=pyo.NonNegativeReals, doc='allocation to player i')
    model.nu = pyo.Var(within=pyo.Reals, doc='nu parameter for minmax core')
    model.eps = pyo.Var(within=pyo.Reals, doc='epsilon parameter for least core')
    # ##########################
    # PARAMETERS
    # ##########################
    model.value = pyo.Param(model.COAL, mutable=True, doc='value of each coalition')
    model.gc = pyo.Param(mutable=True, doc='value of grand coalition')
    # ##########################
    # CONSTRAINTS
    # ##########################
    model.INCREASE = pyo.Constraint(model.COAL, rule=increase_constraint_rule)
    model.EFFICIENCY = pyo.Constraint(rule=efficiency_constraint_rule)
    model.EPS_CONST = pyo.Constraint(rule=eps_constraint_rule)
    model.NU_CONST = pyo.Constraint(rule=nu_constraint_rule)
    # ##########################
    # OBJECTIVE FUNCTION
    # ##########################
    model.OBJ1 = pyo.Objective(sense=pyo.minimize, rule=obj1_rule)
    model.OBJ2 = pyo.Objective(sense=pyo.minimize, rule=obj2_rule)
    model.OBJ3 = pyo.Objective(sense=pyo.maximize, rule=obj3_rule)
    return model

def increase_constraint_rule(model, i):
    return sum(model.xj[j] for j in model.PLAYER_SUB[i]) >= model.nu * model.value[i] - model.eps

def efficiency_constraint_rule(model):
    return sum(model.xj[j] for j in model.PLAYER) == model.gc

def eps_constraint_rule(model):
    return model.eps == 0

def nu_constraint_rule(model):
    return model.nu == 1

def obj1_rule(model):
    return sum(model.xj[j] for j in model.PLAYER)

def obj2_rule(model):
    return model.eps

def obj3_rule(model):
    return model.nu

def instantiate_model(model_data):
    """
    Builds an instance of the optimization model with specific data and objective function.

    Parameters:
    model_data (dict): Dictionary with parameters that populate the model

    Returns:
    pyomo.ConcreteModel: The instantiated model instance
    """
    model = feasible_model()
    problem = model.create_instance(model_data, report_timing=False)
    return problem

def solve_model(instance, GAMS=False):
    """
    Solves the instance of the optimization model.

    Parameters:
    instance (pyomo.ConcreteModel): The model instance to solve
    GAMS (bool, optional): Indicates whether to use GAMS solver. Defaults to False.

    Returns:
    tuple: A tuple containing the solver results and the solved instance
    """
    solver = pyo.SolverFactory('glpk') if not GAMS else pyo.SolverFactory('gams')
    results = solver.solve(instance, keepfiles=False, tee=False, report_timing=False)
    instance.solutions.load_from(results)
    return results, instance

def create_model_data(game):
    """
    Creates model data from the game.

    Parameters:
    game (pd.DataFrame): DataFrame containing the game data

    Returns:
    dict: Dictionary containing the model data
    """
    game['coalition'] = [tuple(x) for x in game['coalition']]
    game = game.set_index('coalition')
    PLAYER = list(set([element for sublist in [i for i in game.index] for element in sublist]))
    COAL = [str(coal) for coal in game.index if str(coal) and str(coal) not in ['()'] and len(coal) < len(PLAYER)]
    GRAND_COAL = [coal for coal in game.index if str(coal) and str(coal) not in ['()'] and len(coal) == len(PLAYER)]
    PLAYER_SUB = {str(coal): list(eval(coal)) for coal in COAL}
    value = {str(coal): game['value'][eval(coal)] for coal in COAL}
    gc = {None: float(game['value'][GRAND_COAL].iloc[0])}
    model_data = {None: {
        'COAL': COAL,
        'PLAYER': PLAYER,
        'PLAYER_SUB': PLAYER_SUB,
        'value': value,
        'gc': gc,
    }}
    return model_data

def core_exists(game, GAMS=False):
    """
    Checks if the set of core allocations is not empty.

    Parameters:
    game (pd.DataFrame): DataFrame containing the game data
    GAMS (bool, optional): Indicates whether to use GAMS solver. Defaults to False.

    Returns:
    bool: True if the core exists, False otherwise
    """
    model_data = create_model_data(game)
    instance = instantiate_model(model_data)
    instance.OBJ1.activate()
    instance.OBJ2.deactivate()
    instance.OBJ3.deactivate()
    results, instance_solved = solve_model(instance, GAMS)
    if results.solver.status == SolverStatus.ok and results.solver.termination_condition == TerminationCondition.optimal:
        print('The set of core allocations is NOT EMPTY\n')
        return True
    else:
        print('WARNING: The set of core allocations is EMPTY (!!!)\n')
        return False

def is_in_core(game, x, eps=0.000000001):
    """
    Checks if a given allocation is in the core.

    Parameters:
    game (pd.DataFrame): DataFrame containing the game data
    x (dict): The allocation dictionary
    eps (float, optional): Tolerance for comparing values. Defaults to 0.000000001.

    Returns:
    bool: True if the allocation is in the core, False otherwise
    """
    ratio = sum(x[p] for p in x) / max(game['value'])
    if ratio < 1 - eps or ratio > 1 + eps:
        return False
    for ind in game.index:
        if sum(x[p] for p in game['coalition'][ind]) < game['value'][ind] * (1 - eps):
            return False
    return True

def least_core(game, GAMS=False):
    """
    Calculates the least core of the game.

    Parameters:
    game (pd.DataFrame): DataFrame containing the game data
    GAMS (bool, optional): Indicates whether to use GAMS solver. Defaults to False.

    Returns:
    dict: Dictionary containing the allocations in the least core
    """
    model_data = create_model_data(game)
    instance = instantiate_model(model_data)
    instance.OBJ1.deactivate()
    instance.OBJ2.activate()
    instance.OBJ3.deactivate()
    instance.EPS_CONST.deactivate()
    results, instance_solved = solve_model(instance, GAMS)
    eps = instance_solved.OBJ2()
    if eps <= 0:
        print('The core is not empty and epsilon is ' + str(eps) + '\n')
    else:
        print('The core is empty and epsilon is ' + str(eps) + '\n')
    return {j: instance_solved.xj[j].value for j in instance.PLAYER}

def minmax_core(game, GAMS=False):
    """
    Calculates the minmax core of the game.

    Parameters:
    game (pd.DataFrame): DataFrame containing the game data
    GAMS (bool, optional): Indicates whether to use GAMS solver. Defaults to False.

    Returns:
    dict: Dictionary containing the allocations in the minmax core
    """
    model_data = create_model_data(game)
    instance = instantiate_model(model_data)
    instance.OBJ1.deactivate()
    instance.OBJ2.deactivate()
    instance.OBJ3.activate()
    instance.NU_CONST.deactivate()
    results, instance_solved = solve_model(instance, GAMS)
    nu = instance_solved.OBJ3()
    if nu >= 1:
        print('The core is not empty and nu is ' + str(nu) + '\n')
    else:
        print('The core is empty and nu is ' + str(nu) + '\n')
    return {j: instance_solved.xj[j].value for j in instance.PLAYER}



import math

def dif_gain(name, n, game):
    """
    Calculates the difference between the gains of coalitions with n players including the player 'name'
    and those of coalitions with n-1 players not containing the player 'name'.

    Parameters:
    name (str): Name of the player
    n (int): Number of players in the coalitions
    game (pd.DataFrame): DataFrame containing the game data

    Returns:
    float: Difference in gains
    """
    sum_in = 0  # Sum of the gains of n-player coalitions including the player 'name'
    sum_ex = 0  # Sum of the gains of n-1 player coalitions not containing the player 'name'

    for ind in game.index:
        coalition = game['coalition'][ind]
        value = game['value'][ind]

        if len(coalition) == n - 1 and name not in coalition:
            sum_ex += value
        elif len(coalition) == n and name in coalition:
            sum_in += value

    return sum_in - sum_ex

def Shapley(game):
    """
    Calculates the Shapley value for each player in a cooperative game.

    Parameters:
    game (pd.DataFrame): DataFrame containing the game data

    Returns:
    dict: Dictionary containing the Shapley value for each player
    """
    players = set([element for sublist in game['coalition'] for element in sublist])
    Shapley = {player: 0 for player in players}

    n = len(players)
    for player in players:
        for j in range(1, n + 1):
            coef = math.factorial(n - j) * math.factorial(j - 1) / math.factorial(n)
            Shapley[player] += coef * dif_gain(player, j, game)

    print('Shapley value properly calculated\n')
    return Shapley



import math

def validate(game):
    """
    Validates the style of the passed dataframe for a cooperative game.

    Parameters:
    game (pd.DataFrame): DataFrame containing the game data

    Returns:
    None
    """
    # Check format of dataframe
    if list(game.columns) == ['coalition', 'value']:
        print('The passed dataframe has the correct columns\n')

    # Check if indices are correct type
    if game.index.dtype == 'int64':
        print('The indices have the correct type (int)\n')

    # Check if coalitions are correct type
    if all(isinstance(coal, list) for coal in game['coalition']):
        print('The coalitions have the correct type (list)\n')

    # Check if values are correct type
    if all(isinstance(value, float) for value in game['value']):
        print('The values have the correct type (float)\n')

    # Number of players according to the number of coalitions
    num_coalitions = len(game)
    num_players = math.log2(num_coalitions)

    # Number of individual players passed in the dataframe
    unique_players = set([element for sublist in game['coalition'] for element in sublist])
    num_individual_players = len(unique_players)

    if num_players == num_individual_players:
        print(f'The number of coalitions ({num_coalitions}) correctly matches the number of individual players in the game ({num_individual_players})\n')

    # Check empty coalition and grand coalition
    empty_coalition = game['coalition'].apply(lambda x: len(x) == 0)
    grand_coalition = game['coalition'].apply(lambda x: len(x) == num_players)

    if (empty_coalition & (game['value'] == 0)).any():
        print('The value of the empty coalition is 0\n')

    if (grand_coalition & (game['value'] == game['value'].max())).any():
        print('The value of the grand coalition is the maximum\n')



import pyomo.environ as pyo
from pyomo.opt import SolverStatus, TerminationCondition
import pandas as pd


def calculate_weights(v, N, players):
    W = {}
    for c in players:
        W_c = 0
        for z in v.keys():
            if c in z:
                subset_without_c = frozenset(z - {c})
                if subset_without_c in v:
                    W_c += v[z] - v[subset_without_c]
                else:
                    print(f"Warning: subset {subset_without_c} not found in v.")
        W[c] = W_c / v[N]
    return W


def create_maali_model(players, v, W, N):
    model = pyo.ConcreteModel()

    # Variables
    model.x = pyo.Var(players, within=pyo.NonNegativeReals)
    model.lam = pyo.Var(within=pyo.NonNegativeReals)

    # Objective
    model.obj = pyo.Objective(expr=model.lam, sense=pyo.maximize)

    # Constraints
    model.constraints = pyo.ConstraintList()

    for c in players:
        if W[c] != 0:
            model.constraints.add((1 / W[c]) * model.x[c] >= model.lam)
        else:
            model.constraints.add(model.x[c] == 0)

    for c in players:
        model.constraints.add(model.x[c] >= v[frozenset({c})])

    model.constraints.add(sum(model.x[c] for c in players) == v[N])

    return model


def solve_maali_model(instance, GAMS=False):
    if GAMS:
        solver = pyo.SolverFactory('gams')
    else:
        solver = pyo.SolverFactory('glpk')
    results = solver.solve(instance, keepfiles=False, tee=False, report_timing=False)
    instance.solutions.load_from(results)
    return results, instance


def maali(game, GAMS=False):
    # Prepare the data
    #game['coalition'] = game['coalition'].apply(frozenset)
    v = {frozenset(coalition): value for coalition, value in zip(game['coalition'], game['value'])}

    players = list(set(player for coalition in game['coalition'] for player in coalition))
    N = frozenset(players)

    # Calculate weights
    W = calculate_weights(v, N, players)

    # Create and solve the model
    model = create_maali_model(players, v, W, N)
    results, solved_model = solve_maali_model(model, GAMS)

    # Return the results
    if (results.solver.status == SolverStatus.ok) and (
            results.solver.termination_condition == TerminationCondition.optimal):
        print("Maali's solution properly calculated.")
        return {j: pyo.value(solved_model.x[j]) for j in players}
    else:
        print("Solver failed to find an optimal solution.")
        return None




import random
from itertools import combinations
import pandas as pd

def create_game(n=1, seed=None):
    '''
    Function that creates a (random) n-player TU game
    :param n: number of players in the game
    :param seed: seed for random number generation (optional)
    :return: A fully specified game in the form of a DataFrame
    '''
    # Set random seed if provided
    if seed is not None:
        random.seed(seed)

    players = range(n)
    # Generate all coalitions for the n-player game

    coalitions = [comb for i in range(n+1) for comb in combinations(players, i)]

    # Generate benefits for all coalitions
    values = {(): 0.0}
    game = {'coalition': [], 'value': []}
    for k, coal in enumerate(coalitions, start=0):
        incr = random.random()
        pre = max(values[tuple([x for x in coal if x != player])] for player in coal) if len(coal) > 0 else 0
        if len(coal) <= 1:
            values[coal] = 0
            game['coalition'].append(list(coal))
            game['value'].append(0)
        else:
            values[coal] = pre + incr
            game['coalition'].append(list(coal))
            game['value'].append(pre + incr)

    game_df = pd.DataFrame(game)
    print(f'The {n}-player game has been created')
    print('###############\n')
    return game_df



from setuptools import setup, find_packages

# Open and read the contents of the README file
with open('README.md', 'r', encoding='utf-8') as readme_file:
    long_description = readme_file.read()

setup(
    name="pyCoopGame",
    version="0.0.6",
    description="various profit allocation methods from cooperative game theory",
    author="Fabian Lechtenberg",
    author_email="fabian.lechtenberg@upc.edu",
    url="https://github.com/flechtenberg/pyCoopGame",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    packages=find_packages(),
    python_requires=">=3.9, <=3.12",
    install_requires=[
    ],
    long_description=long_description,
    long_description_content_type='text/markdown',  # Specify the content type of the README
)



import pyomo.environ as pyo
from pyomo.opt import SolverStatus, TerminationCondition

def create_model():
    """ This function builds an abstract model on top of the ecoinvent database """
    model = pyo.AbstractModel()
    # SETS
    model.COAL = pyo.Set(doc='Set of coalitions')
    model.INDEX = pyo.Set(doc='Set of indices')
    model.FINAL_INDEX = pyo.Set(within=model.INDEX)
    model.PLAYER = pyo.Set(doc='Set of players')
    model.PLAYER_SUB = pyo.Set(model.COAL, doc='Player subsets')
    # VARIABLES
    model.dik = pyo.Var(model.COAL, model.INDEX - model.FINAL_INDEX, within=pyo.NonNegativeReals, doc='dik')
    model.tk = pyo.Var(model.INDEX - model.FINAL_INDEX, within=pyo.Reals)
    model.phi = pyo.Var(model.COAL, within=pyo.Reals)
    model.xj = pyo.Var(model.PLAYER, within=pyo.NonNegativeReals)
    # PARAMETERS
    model.indexk = pyo.Param(model.INDEX, mutable=True, within=pyo.Reals)
    model.lambdak = pyo.Param(model.INDEX, mutable=True, within=pyo.Reals, doc='lambda')
    model.value = pyo.Param(model.COAL, mutable=True, within=pyo.Reals)
    model.gc = pyo.Param(mutable=True, within=pyo.Reals)
    # CONSTRAINTS
    model.ORDER = pyo.Constraint(model.COAL, model.INDEX - model.FINAL_INDEX,
                                 rule=lambda model, i, k: model.dik[i, k] >= model.phi[i] - model.tk[k])
    model.EXCESS = pyo.Constraint(model.COAL, rule=lambda model, i: model.phi[i] == model.value[i] - sum(model.xj[j] for j in model.PLAYER_SUB[i]))
    model.EFFICIENCY = pyo.Constraint(rule=lambda model: sum(model.xj[j] for j in model.PLAYER) == model.gc)
    # OBJECTIVE
    model.OBJ = pyo.Objective(sense=pyo.minimize,
                              rule=lambda model: sum((model.lambdak[k] - model.lambdak[k + 1]) * (model.indexk[k] * model.tk[k] + sum(model.dik[i, k] for i in model.COAL)) for k in model.INDEX - model.FINAL_INDEX))
    return model

def instantiate(model_data):
    """ This function builds an instance of the optimization model with specific data and objective function

    Parameters:
    model_data: Dictionary with parameters that populate the model
    """
    model = create_model()
    problem = model.create_instance(model_data, report_timing=False)
    return problem

def solve_model(instance, GAMS=False):
    """ This function solves the instance of the optimization model

    Parameters:
    GAMS: Indicates whether to use GAMS solver
    """
    # Solver set-up
    if GAMS is False:
        solver = pyo.SolverFactory('glpk')
    else:
        solver = pyo.SolverFactory('gams')
    results = solver.solve(instance, keepfiles=False, tee=False, report_timing=False) #, io_options=io_options)
    instance.solutions.load_from(results)
    return results, instance

def nucleolus(game, delta=0.5, GAMS=False, tk=False):
    COAL = [str(coal) for coal in game['coalition']]
    INDEX = game.index.tolist()
    FINAL_INDEX = [len(COAL) - 1]
    PLAYER = list(set([element for sublist in game['coalition'] for element in sublist]))
    PLAYER_SUB = {str(coal): coal for coal in game['coalition']}

    indexk = {k: k for k in INDEX}
    lambdak = {k: delta ** (k - 1) for k in INDEX}
    lambdak[FINAL_INDEX[0]] = 0
    value = {str(game['coalition'][ind]): game['value'][ind] for ind in game.index}
    gc = {None: max(game['value'])}

    model_data = {
        None: {
            'COAL': COAL,
            'INDEX': INDEX,
            'FINAL_INDEX': FINAL_INDEX,
            'PLAYER': PLAYER,
            'PLAYER_SUB': PLAYER_SUB,
            'indexk': indexk,
            'lambdak': lambdak,
            'value': value,
            'gc': gc,
        }
    }

    instance = instantiate(model_data)

    results, instance_solved = solve_model(instance, GAMS)
    if tk:
        tks = [instance_solved.tk[j].value for j in [ind for ind in INDEX if ind not in FINAL_INDEX]]
        xjs = {j: instance_solved.xj[j].value for j in PLAYER}
        print(tks)
        print(xjs)
        return tks

    if (results.solver.status == SolverStatus.ok) and (results.solver.termination_condition == TerminationCondition.optimal):
        print('Nucleolus properly calculated\n')
    else:
        print('Solver failed while calculating Nucleolus!\n')
    return {j: instance_solved.xj[j].value for j in PLAYER}

def compare_tks(tk1, tk2):
    if tk1 == tk2:
        print('equal')
        return 0
    print('not equal')
    for t1, t2 in zip(tk1, tk2):
        if t1 > t2:
            return -1
        if t2 > t1:
            return 1
    return 0

def nucl_screen(game, deltas, GAMS=False):
    tks = {}
    ranking = {}
    for delta in deltas:
        print(delta)
        tks[delta] = nucleolus(game, delta=delta, GAMS=GAMS, tk=True)
        ranking[delta] = 0
    for delta1 in deltas:
        for delta2 in deltas:
            incr = compare_tks(tks[delta1], tks[delta2])
            ranking[delta1] -= incr
            ranking[delta2] += incr
    print(ranking)
